@@ -6,18 +6,18 @@
  *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  This library is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU Lesser General Public
+ *  License as published by the Free Software Foundation; either
+ *  version 2.1 of the License, or (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
+ *  You should have received a copy of the GNU Lesser General Public
+ *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
@@ -31,15 +31,25 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "bt.h"
+#include "monitor/bt.h"
 #include "btdev.h"
 
 #define le16_to_cpu(val) (val)
+#define le32_to_cpu(val) (val)
 #define cpu_to_le16(val) (val)
+#define cpu_to_le32(val) (val)
+
+#define has_bredr(btdev)	(!((btdev)->features[4] & 0x20))
+#define has_le(btdev)		(!!((btdev)->features[4] & 0x40))
 
 struct btdev {
+	enum btdev_type type;
+
 	struct btdev *conn;
 
+	btdev_command_func command_handler;
+	void *command_data;
+
 	btdev_send_func send_handler;
 	void *send_data;
 
@@ -160,30 +170,18 @@ static void hexdump(const unsigned char *buf, uint16_t len)
 	}
 }
 
-static void get_bdaddr(uint16_t id, uint8_t *bdaddr)
+static void get_bdaddr(uint16_t id, uint8_t index, uint8_t *bdaddr)
 {
 	bdaddr[0] = id & 0xff;
 	bdaddr[1] = id >> 8;
-	bdaddr[2] = 0x00;
+	bdaddr[2] = index;
 	bdaddr[3] = 0x01;
 	bdaddr[4] = 0xaa;
 	bdaddr[5] = 0x00;
 }
 
-struct btdev *btdev_create(uint16_t id)
+static void set_bredr_features(struct btdev *btdev)
 {
-	struct btdev *btdev;
-
-	btdev = malloc(sizeof(*btdev));
-	if (!btdev)
-		return NULL;
-
-	memset(btdev, 0, sizeof(*btdev));
-
-	btdev->manufacturer = 63;
-	btdev->version = 0x06;
-	btdev->revision = 0x0000;
-
 	btdev->features[0] |= 0x04;	/* Encryption */
 	btdev->features[0] |= 0x20;	/* Role switch */
 	btdev->features[0] |= 0x80;	/* Sniff mode */
@@ -206,15 +204,59 @@ struct btdev *btdev_create(uint16_t id)
 	btdev->features[7] |= 0x01;	/* Link Supervision Timeout Event */
 	btdev->features[7] |= 0x02;	/* Inquiry TX Power Level */
 	btdev->features[7] |= 0x80;	/* Extended features */
+}
+
+static void set_le_features(struct btdev *btdev)
+{
+	btdev->features[4] |= 0x20;	/* BR/EDR Not Supported */
+	btdev->features[4] |= 0x40;	/* LE Supported */
+	btdev->features[7] |= 0x80;	/* Extended features */
+}
+
+static void set_amp_features(struct btdev *btdev)
+{
+}
+
+struct btdev *btdev_create(enum btdev_type type, uint16_t id)
+{
+	struct btdev *btdev;
+	int index;
+
+	btdev = malloc(sizeof(*btdev));
+	if (!btdev)
+		return NULL;
+
+	memset(btdev, 0, sizeof(*btdev));
+	btdev->type = type;
+
+	btdev->manufacturer = 63;
+	btdev->version = 0x06;
+	btdev->revision = 0x0000;
+
+	switch (btdev->type) {
+	case BTDEV_TYPE_BREDR:
+		set_bredr_features(btdev);
+		break;
+	case BTDEV_TYPE_LE:
+		set_le_features(btdev);
+		break;
+	case BTDEV_TYPE_AMP:
+		set_amp_features(btdev);
+		break;
+	}
 
 	btdev->acl_mtu = 192;
 	btdev->acl_max_pkt = 1;
 
 	btdev->country_code = 0x00;
 
-	get_bdaddr(id, btdev->bdaddr);
+	index = add_btdev(btdev);
+	if (index < 0) {
+		free(btdev);
+		return NULL;
+	}
 
-	add_btdev(btdev);
+	get_bdaddr(id, index, btdev->bdaddr);
 
 	return btdev;
 }
@@ -229,6 +271,24 @@ void btdev_destroy(struct btdev *btdev)
 	free(btdev);
 }
 
+void btdev_set_bdaddr(struct btdev *btdev, uint8_t *bdaddr)
+{
+	if (!btdev)
+		return;
+
+	memcpy(btdev->bdaddr, bdaddr, 6);
+}
+
+void btdev_set_command_handler(struct btdev *btdev, btdev_command_func handler,
+							void *user_data)
+{
+	if (!btdev)
+		return;
+
+	btdev->command_handler = handler;
+	btdev->command_data = user_data;
+}
+
 void btdev_set_send_handler(struct btdev *btdev, btdev_send_func handler,
 							void *user_data)
 {
@@ -349,6 +409,8 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 
 			ir.num_resp = 0x01;
 			memcpy(ir.bdaddr, btdev_list[i]->bdaddr, 6);
+			ir.pscan_rep_mode = 0x00;
+			ir.pscan_period_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
 			ir.rssi = -60;
 			memcpy(ir.data, btdev_list[i]->ext_inquiry_rsp, 240);
@@ -363,6 +425,8 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 
 			ir.num_resp = 0x01;
 			memcpy(ir.bdaddr, btdev_list[i]->bdaddr, 6);
+			ir.pscan_rep_mode = 0x00;
+			ir.pscan_period_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
 			ir.rssi = -60;
 
@@ -373,6 +437,9 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 
 			ir.num_resp = 0x01;
 			memcpy(ir.bdaddr, btdev_list[i]->bdaddr, 6);
+			ir.pscan_rep_mode = 0x00;
+			ir.pscan_period_mode = 0x00;
+			ir.pscan_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
 
 			send_event(btdev, BT_HCI_EVT_INQUIRY_RESULT,
@@ -471,7 +538,7 @@ static void disconnect_complete(struct btdev *btdev, uint16_t handle,
 static void name_request_complete(struct btdev *btdev,
 					const uint8_t *bdaddr, uint8_t status)
 {
-        struct bt_hci_evt_remote_name_req_complete nc;
+        struct bt_hci_evt_remote_name_request_complete nc;
 
 	nc.status = status;
 	memcpy(nc.bdaddr, bdaddr, 6);
@@ -566,9 +633,9 @@ static void remote_version_complete(struct btdev *btdev, uint16_t handle)
 							&rvc, sizeof(rvc));
 }
 
-static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
+static void default_cmd(struct btdev *btdev, uint16_t opcode,
+						const void *data, uint8_t len)
 {
-	const struct bt_hci_cmd_hdr *hdr = data;
 	const struct bt_hci_cmd_create_conn *cc;
 	const struct bt_hci_cmd_disconnect *dc;
 	const struct bt_hci_cmd_create_conn_cancel *ccc;
@@ -591,7 +658,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 	const struct bt_hci_cmd_write_voice_setting *wvs;
 	const struct bt_hci_cmd_write_inquiry_mode *wim;
 	const struct bt_hci_cmd_write_afh_assess_mode *waam;
-	const struct bt_hci_cmd_write_ext_inquiry_rsp *weir;
+	const struct bt_hci_cmd_write_ext_inquiry_response *weir;
 	const struct bt_hci_cmd_write_simple_pairing_mode *wspm;
 	const struct bt_hci_cmd_write_le_host_supported *wlhs;
 	const struct bt_hci_cmd_le_set_event_mask *lsem;
@@ -608,9 +675,9 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 	struct bt_hci_rsp_read_voice_setting rvs;
 	struct bt_hci_rsp_read_inquiry_mode rim;
 	struct bt_hci_rsp_read_afh_assess_mode raam;
-	struct bt_hci_rsp_read_ext_inquiry_rsp reir;
+	struct bt_hci_rsp_read_ext_inquiry_response reir;
 	struct bt_hci_rsp_read_simple_pairing_mode rspm;
-	struct bt_hci_rsp_read_inquiry_rsp_tx_power rirtp;
+	struct bt_hci_rsp_read_inquiry_resp_tx_power rirtp;
 	struct bt_hci_rsp_read_le_host_supported rlhs;
 	struct bt_hci_rsp_read_local_version rlv;
 	struct bt_hci_rsp_read_local_commands rlc;
@@ -620,17 +687,13 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 	struct bt_hci_rsp_read_country_code rcc;
 	struct bt_hci_rsp_read_bd_addr rba;
 	struct bt_hci_rsp_read_data_block_size rdbs;
+	struct bt_hci_rsp_read_local_amp_info rlai;
 	struct bt_hci_rsp_le_read_buffer_size lrbs;
 	struct bt_hci_rsp_le_read_local_features lrlf;
+	struct bt_hci_rsp_le_read_adv_tx_power lratp;
 	struct bt_hci_rsp_le_read_supported_states lrss;
-	uint16_t opcode;
 	uint8_t status, page;
 
-	if (len < sizeof(*hdr))
-		return;
-
-	opcode = le16_to_cpu(hdr->opcode);
-
 	switch (opcode) {
 	case BT_HCI_CMD_INQUIRY:
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
@@ -643,43 +706,43 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_CREATE_CONN:
-		cc = data + sizeof(*hdr);
+		cc = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		conn_request(btdev, cc->bdaddr);
 		break;
 
 	case BT_HCI_CMD_DISCONNECT:
-		dc = data + sizeof(*hdr);
+		dc = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		disconnect_complete(btdev, le16_to_cpu(dc->handle), dc->reason);
 		break;
 
 	case BT_HCI_CMD_CREATE_CONN_CANCEL:
-		ccc = data + sizeof(*hdr);
+		ccc = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		conn_complete(btdev, ccc->bdaddr, BT_HCI_ERR_UNKNOWN_CONN_ID);
 		break;
 
 	case BT_HCI_CMD_ACCEPT_CONN_REQUEST:
-		acr = data + sizeof(*hdr);
+		acr = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		conn_complete(btdev, acr->bdaddr, BT_HCI_ERR_SUCCESS);
 		break;
 
 	case BT_HCI_CMD_REJECT_CONN_REQUEST:
-		rcr = data + sizeof(*hdr);
+		rcr = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		conn_complete(btdev, rcr->bdaddr, BT_HCI_ERR_UNKNOWN_CONN_ID);
 		break;
 
 	case BT_HCI_CMD_REMOTE_NAME_REQUEST:
-		rnr = data + sizeof(*hdr);
+		rnr = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		name_request_complete(btdev, rnr->bdaddr, BT_HCI_ERR_SUCCESS);
 		break;
 
 	case BT_HCI_CMD_REMOTE_NAME_REQUEST_CANCEL:
-		rnrc = data + sizeof(*hdr);
+		rnrc = data;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
 		name_request_complete(btdev, rnrc->bdaddr,
@@ -687,20 +750,20 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_READ_REMOTE_FEATURES:
-		rrf = data + sizeof(*hdr);
+		rrf = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		remote_features_complete(btdev, le16_to_cpu(rrf->handle));
 		break;
 
 	case BT_HCI_CMD_READ_REMOTE_EXT_FEATURES:
-		rref = data + sizeof(*hdr);
+		rref = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		remote_ext_features_complete(btdev, le16_to_cpu(rref->handle),
 								rref->page);
 		break;
 
 	case BT_HCI_CMD_READ_REMOTE_VERSION:
-		rrv = data + sizeof(*hdr);
+		rrv = data;
 		cmd_status(btdev, BT_HCI_ERR_SUCCESS, opcode);
 		remote_version_complete(btdev, le16_to_cpu(rrv->handle));
 		break;
@@ -712,14 +775,14 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_DEFAULT_LINK_POLICY:
-		wdlp = data + sizeof(*hdr);
+		wdlp = data;
 		btdev->default_link_policy = le16_to_cpu(wdlp->policy);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
 		break;
 
 	case BT_HCI_CMD_SET_EVENT_MASK:
-		sem = data + sizeof(*hdr);
+		sem = data;
 		memcpy(btdev->event_mask, sem->mask, 8);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -731,7 +794,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_SET_EVENT_FILTER:
-		sef = data + sizeof(*hdr);
+		sef = data;
 		btdev->event_filter = sef->type;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -757,7 +820,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_LOCAL_NAME:
-		wln = data + sizeof(*hdr);
+		wln = data;
 		memcpy(btdev->name, wln->name, 248);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -776,7 +839,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_CONN_ACCEPT_TIMEOUT:
-		wcat = data + sizeof(*hdr);
+		wcat = data;
 		btdev->conn_accept_timeout = le16_to_cpu(wcat->timeout);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -789,7 +852,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_PAGE_TIMEOUT:
-		wpt = data + sizeof(*hdr);
+		wpt = data;
 		btdev->page_timeout = le16_to_cpu(wpt->timeout);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -802,7 +865,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_SCAN_ENABLE:
-		wse = data + sizeof(*hdr);
+		wse = data;
 		btdev->scan_enable = wse->enable;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -815,7 +878,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_AUTH_ENABLE:
-		wae = data + sizeof(*hdr);
+		wae = data;
 		btdev->auth_enable = wae->enable;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -828,7 +891,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_CLASS_OF_DEV:
-		wcod = data + sizeof(*hdr);
+		wcod = data;
 		memcpy(btdev->dev_class, wcod->dev_class, 3);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -841,7 +904,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_VOICE_SETTING:
-		wvs = data + sizeof(*hdr);
+		wvs = data;
 		btdev->voice_setting = le16_to_cpu(wvs->setting);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -854,7 +917,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_INQUIRY_MODE:
-		wim = data + sizeof(*hdr);
+		wim = data;
 		btdev->inquiry_mode = wim->mode;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -867,21 +930,21 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_AFH_ASSESS_MODE:
-		waam = data + sizeof(*hdr);
+		waam = data;
 		btdev->afh_assess_mode = waam->mode;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
 		break;
 
-	case BT_HCI_CMD_READ_EXT_INQUIRY_RSP:
+	case BT_HCI_CMD_READ_EXT_INQUIRY_RESPONSE:
 		reir.status = BT_HCI_ERR_SUCCESS;
 		reir.fec = btdev->ext_inquiry_fec;
 		memcpy(reir.data, btdev->ext_inquiry_rsp, 240);
 		cmd_complete(btdev, opcode, &reir, sizeof(reir));
 		break;
 
-	case BT_HCI_CMD_WRITE_EXT_INQUIRY_RSP:
-		weir = data + sizeof(*hdr);
+	case BT_HCI_CMD_WRITE_EXT_INQUIRY_RESPONSE:
+		weir = data;
 		btdev->ext_inquiry_fec = weir->fec;
 		memcpy(btdev->ext_inquiry_rsp, weir->data, 240);
 		status = BT_HCI_ERR_SUCCESS;
@@ -895,13 +958,13 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_SIMPLE_PAIRING_MODE:
-		wspm = data + sizeof(*hdr);
+		wspm = data;
 		btdev->simple_pairing_mode = wspm->mode;
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
 		break;
 
-	case BT_HCI_CMD_READ_INQUIRY_RSP_TX_POWER:
+	case BT_HCI_CMD_READ_INQUIRY_RESP_TX_POWER:
 		rirtp.status = BT_HCI_ERR_SUCCESS;
 		rirtp.level = 0;
 		cmd_complete(btdev, opcode, &rirtp, sizeof(rirtp));
@@ -915,7 +978,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_WRITE_LE_HOST_SUPPORTED:
-		wlhs = data + sizeof(*hdr);
+		wlhs = data;
 		btdev->le_supported = wlhs->supported;
 		btdev->le_simultaneous = wlhs->simultaneous;
 		status = BT_HCI_ERR_SUCCESS;
@@ -945,7 +1008,7 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		break;
 
 	case BT_HCI_CMD_READ_LOCAL_EXT_FEATURES:
-		page = ((const uint8_t *) data)[sizeof(*hdr)];
+		page = ((const uint8_t *) data)[0];
 		switch (page) {
 		case 0x00:
 			rlef.status = BT_HCI_ERR_SUCCESS;
@@ -1004,8 +1067,23 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		cmd_complete(btdev, opcode, &rdbs, sizeof(rdbs));
 		break;
 
+	case BT_HCI_CMD_READ_LOCAL_AMP_INFO:
+		rlai.status = BT_HCI_ERR_SUCCESS;
+		rlai.amp_status = 0x01;		/* Used for Bluetooth only */
+		rlai.total_bw = cpu_to_le32(0);
+		rlai.max_bw = cpu_to_le32(0);
+		rlai.min_latency = cpu_to_le32(0);
+		rlai.max_pdu = cpu_to_le32(672);
+		rlai.amp_type = 0x01;		/* 802.11 AMP Controller */
+		rlai.pal_cap = cpu_to_le16(0x0000);
+		rlai.max_assoc_len = cpu_to_le16(672);
+		rlai.max_flush_to = cpu_to_le32(0xffffffff);
+		rlai.be_flush_to = cpu_to_le32(0xffffffff);
+		cmd_complete(btdev, opcode, &rlai, sizeof(rlai));
+		break;
+
 	case BT_HCI_CMD_LE_SET_EVENT_MASK:
-		lsem = data + sizeof(*hdr);
+		lsem = data;
 		memcpy(btdev->le_event_mask, lsem->mask, 8);
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -1024,6 +1102,12 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 		cmd_complete(btdev, opcode, &lrlf, sizeof(lrlf));
 		break;
 
+	case BT_HCI_CMD_LE_READ_ADV_TX_POWER:
+		lratp.status = BT_HCI_ERR_SUCCESS;
+		lratp.level = 0;
+		cmd_complete(btdev, opcode, &lratp, sizeof(lratp));
+		break;
+
 	case BT_HCI_CMD_LE_SET_SCAN_PARAMETERS:
 		status = BT_HCI_ERR_SUCCESS;
 		cmd_complete(btdev, opcode, &status, sizeof(status));
@@ -1048,6 +1132,67 @@ static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
 	}
 }
 
+struct btdev_callback {
+	void (*function)(btdev_callback callback, uint8_t response,
+				uint8_t status, const void *data, uint8_t len);
+	void *user_data;
+	uint16_t opcode;
+	const void *data;
+	uint8_t len;
+};
+
+void btdev_command_response(btdev_callback callback, uint8_t response,
+                                uint8_t status, const void *data, uint8_t len)
+{
+	callback->function(callback, response, status, data, len);
+}
+
+static void handler_callback(btdev_callback callback, uint8_t response,
+				uint8_t status, const void *data, uint8_t len)
+{
+	struct btdev *btdev = callback->user_data;
+
+	switch (response) {
+	case BTDEV_RESPONSE_DEFAULT:
+		default_cmd(btdev, callback->opcode,
+					callback->data, callback->len);
+		break;
+	case BTDEV_RESPONSE_COMMAND_STATUS:
+		cmd_status(btdev, status, callback->opcode);
+		break;
+	case BTDEV_RESPONSE_COMMAND_COMPLETE:
+		cmd_complete(btdev, callback->opcode, data, len);
+		break;
+	default:
+		cmd_status(btdev, BT_HCI_ERR_UNKNOWN_COMMAND,
+						callback->opcode);
+		break;
+	}
+}
+
+static void process_cmd(struct btdev *btdev, const void *data, uint16_t len)
+{
+	struct btdev_callback callback;
+	const struct bt_hci_cmd_hdr *hdr = data;
+
+	if (len < sizeof(*hdr))
+		return;
+
+	callback.function = handler_callback;
+	callback.user_data = btdev;
+	callback.opcode = le16_to_cpu(hdr->opcode);
+	callback.data = data + sizeof(*hdr);
+	callback.len = hdr->plen;
+
+	if (btdev->command_handler)
+		btdev->command_handler(callback.opcode,
+					callback.data, callback.len,
+					&callback, btdev->command_data);
+	else
+		default_cmd(btdev, callback.opcode,
+					callback.data, callback.len);
+}
+
 void btdev_receive_h4(struct btdev *btdev, const void *data, uint16_t len)
 {
 	uint8_t pkt_type;
